# Copyright (c) 2003 CORE Security Technologies
#
# This software is provided under under a slightly modified version
# of the Apache Software License. See the accompanying LICENSE file
# for more information.
#
# $Id: __init__.py,v 1.2 2003/10/27 17:36:57 jkohen Exp $

pass
